/* 
 * Copyright (C) 2004  Free Software Initiative of Japan
 *
 * This file is subject to the terms and conditions of the GNU Lesser
 * General Public License version 2.1 (or later).
 *
 */

#include "types.h"
#include "pp.h"
#include "cache.h"
#include "char-io.h"
#include "delay.h"

#include "eth.h"
#include "netinet.h"
#include "netinet6.h"
#include "string.h"

#include "led.h"
#include "serial.h"
#include "ata.h"

#include "conf-machine.h"
#include "conf-cpu.h"
#include "conf-mem.h"

/* Go off! */
static const char * const banner = "\ng00ff version -36.2 with IPv6 (M32R)\n"
	"\tCopyright 2004, 2005  Free Software Initiative of Japan\n";

typedef int (*bioscall) ();

static int
nullcall (void)
{
  return 0;
}

/**
 * machine_reset - Reset the machine
 * @how: cold boot, warm boot, suspend or sleep
 *
 */
static int
machine_reset (int how)
{
  register unsigned int __r0 __asm__ ("r0") = (unsigned int) how;
  register unsigned int __r4 __asm__ ("r4") = (unsigned int) TEXT_START;

  asm volatile ("jmp	%0\n\tnop"
		:
		: "r" (__r4), "r" (__r0));

  /* Never reached */
  return 0;
}

/**
 * get_cpu_info - Get CPU information
 * @cpu_info: Pointer to the information to be filled
 *
 */
static int
get_cpu_info (unsigned long *cpu_info)
{
  *(cpu_info + 0) = CPUCLK;
  *(cpu_info + 1) = BUSCLK;
  *(cpu_info + 2) = TIMER_DIVIDE;

  return 0;
}

bioscall bios_call_vector[]  = {
  put_char, put_string, ata_read_sectors, flush_cache,
  machine_reset, get_cpu_info, nullcall, nullcall,
  nullcall, nullcall, nullcall, nullcall,
  nullcall, nullcall, nullcall, nullcall,
};

extern void _heap;
static unsigned char *buf = &_heap;

static void
g00ff (void)
{
  register unsigned int __r0 __asm__ ("r0") = (unsigned int) bios_call_vector;
  register unsigned int __r1 __asm__ ("r1") = (unsigned int) MEMORY_START;
  register unsigned int __r2 __asm__ ("r2") = (unsigned int) MEMORY_SIZE;
  register unsigned int __r4 __asm__ ("r4") = (unsigned int) buf;

  asm volatile ("jmp	%0\n\tnop"
		:
		: "r" (__r4), "r" (__r0), "r" (__r1), "r" (__r2));
}

static void inline
goto_kernel (unsigned long bios_call_vector,
	     unsigned char *main_mem_start,
	     unsigned long mem_size,
	     unsigned char *initrd_start,
	     unsigned long initrd_size,
	     unsigned char *zimage_address)
{
  unsigned char *kernel_params = (unsigned char *)main_mem_start + 0x1000;

  /* XXX
   * We should specify memory size with mem_size
   */

  /* 0xff000400 is "bra +0x1000", because of wrong firmware vector */
  *(unsigned long *)(kernel_params + 0x0000) = 0xff000400; /* mount rd only */
  *(unsigned long *)(kernel_params + 0x0004) = 0x4000;	/* ramdisk flag */
  *(unsigned long *)(kernel_params + 0x0008) = 0x0301;	/* orig root */
  *(unsigned long *)(kernel_params + 0x000c) = 1;	/* loader type */
  *(unsigned long *)(kernel_params + 0x0010) = (unsigned long)initrd_start;
  *(unsigned long *)(kernel_params + 0x0014) = initrd_size;

  get_cpu_info ((unsigned long *)(kernel_params + 0x0018));

  kernel_params[0x100] = 'c';
  kernel_params[0x101] = 'o';
  kernel_params[0x102] = 'n';
  kernel_params[0x103] = 's';
  kernel_params[0x104] = 'o';
  kernel_params[0x105] = 'l';
  kernel_params[0x106] = 'e';
  kernel_params[0x107] = '=';
  kernel_params[0x108] = 't';
  kernel_params[0x109] = 't';
  kernel_params[0x10a] = 'y';
  kernel_params[0x10b] = '1';
  kernel_params[0x10c] = ' ';
  kernel_params[0x10d] = 'c';
  kernel_params[0x10e] = 'o';
  kernel_params[0x10f] = 'n';
  kernel_params[0x110] = 's';
  kernel_params[0x111] = 'o';
  kernel_params[0x112] = 'l';
  kernel_params[0x113] = 'e';
  kernel_params[0x114] = '=';
  kernel_params[0x115] = 't';
  kernel_params[0x116] = 't';
  kernel_params[0x117] = 'y';
  kernel_params[0x118] = 'D';
  kernel_params[0x119] = '0';
  kernel_params[0x11a] = ',';
  kernel_params[0x11b] = '1';
  kernel_params[0x11c] = '1';
  kernel_params[0x11d] = '5';
  kernel_params[0x11e] = '2';
  kernel_params[0x11f] = '0';
  kernel_params[0x120] = '0';
  kernel_params[0x121] = 'n';
  kernel_params[0x122] = '8';
  kernel_params[0x123] = 'x';
  kernel_params[0x124] = ' ';
  kernel_params[0x125] = 'r';
  kernel_params[0x126] = 'o';
  kernel_params[0x127] = 'o';
  kernel_params[0x128] = 't';
  kernel_params[0x129] = '=';
  kernel_params[0x12a] = '/';
  kernel_params[0x12b] = 'd';
  kernel_params[0x12c] = 'e';
  kernel_params[0x12d] = 'v';
  kernel_params[0x12e] = '/';
  kernel_params[0x12f] = 'h';
  kernel_params[0x130] = 'd';
  kernel_params[0x131] = 'a';
  kernel_params[0x132] = '1';
  kernel_params[0x133] = ' ';
  kernel_params[0x134] = 'm';
  kernel_params[0x135] = 'e';
  kernel_params[0x136] = 'm';
  kernel_params[0x137] = '=';
  kernel_params[0x138] = '3';
  kernel_params[0x139] = '2';
  kernel_params[0x13a] = 'M';
  kernel_params[0x13b] = '\0';
  kernel_params[0x13c] = '\0';
  kernel_params[0x13d] = '\0';
  kernel_params[0x13e] = '\0';
  kernel_params[0x13f] = '\0';

  {
    register long __r0 __asm__ ("r0") = (long)bios_call_vector;
    register long __r1 __asm__ ("r1") = (long)zimage_address;
    register long __r2 __asm__ ("r2") = (long)mem_size;

    asm volatile ("jmp	r1\n\t.fillinsn"
		  : 
		  : "r" (__r0), "r" (__r1), "r" (__r2)
		  : "memory");
  }
}

#define MAX_REQUEST 1024
void
boot (void)
{
  static int x;			/* This is in the BSS segment */
  unsigned short data = x + 1;	/* We can check BSS is zero-ed out or not */
#ifdef NET_BOOT_IPv4
  struct eth_addr eth_addr;
  u32 ip_addr, netmask, router_ip_addr, server_ip_addr;
  u32 dns_server, server;
#else  /* IPv6 */
  int i;
  struct eth_addr *dst_eth_addr;
  /* 2001:218:2001:3000::161 */
  struct ipv6_addr ipv6_server = {{0x20, 0x01, 0x02, 0x18, 0x20, 0x01, 0x30, 0x00,
				   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x61}};
  struct ipv6_addr ipv6_router = {{0x20, 0x01, 0x02, 0x40, 0x01, 0x12, 0x00, 0x04,
				   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01}};
  struct ipv6_addr ipv6_note =  {{0x20, 0x01, 0x02, 0x40, 0x01, 0x12, 0x00, 0x04,
				  0x02, 0x12, 0x3f, 0xff, 0xfe, 0x69, 0x86, 0xd8}};
  /* ns.tokyo.wide.ad.jp */
  struct ipv6_addr dns_server = {{0x20, 0x01, 0x02, 0x00, 0x00, 0x00, 0x00, 0x01, 
				  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08}};
#endif
  struct eth_packet *p = (struct eth_packet *)buf;
  unsigned char *recv_buf = (unsigned char *)MEMORY_START+0x01000000;
  int r;
  u8 http_request[MAX_REQUEST];
  u16 server_port;
  u8 *s;
  u8 *server_name = SERVER_NAME;
  u8 *path = KERNEL_PATH;

  serial_init ();
  put_string (banner);

  if (ata_init () < 0)
    {
      put_string ("*** ATA initialization failed.\n");
      for (;;);
    }

#ifdef NET_BOOT
  if (eth_init (p) < 0)
    {
      put_string ("*** Ethernet initialization failed.\n");
      for (;;);
    }
  else
      put_string (", Ethernet initialization done.\n");

#ifdef NET_BOOT_IPv4
  inet_aton ("192.168.2.124", &ip_addr); /* This machine's IP address */
  inet_aton ("255.255.255.0", &netmask); /* This machine's netmask */
  inet_aton ("192.168.2.1", &dns_server); /* DNS server address */
  inet_aton ("192.168.2.1", &router_ip_addr); /* Router's IP address */

#if defined(DHCP_SETUP)
  {
    struct dhcp_data dd;

    if (dhcp (p, &dd) < 0)
      {
	put_string ("DHCP failed.\n");
	for (;;);
      }

    put_string ("DHCP success.\n");

    ip_addr = dd.my_ip_addr;
    if (dd.netmask)
      netmask = dd.netmask;
    if (dd.router_ip_addr)
      router_ip_addr = dd.router_ip_addr;
    if (dd.dns)
      dns_server = dd.dns;
    if (dd.server_name != 0)
      server_name = dd.server_name;
    if (dd.filename != 0)
      path = dd.filename;
  }
#endif

  net_interface_config (ip_addr, netmask);

  if ((netmask & ip_addr) == (netmask & dns_server))
    server = dns_server;
  else
    server = router_ip_addr;

  if (arp_resolv (p, server, &eth_addr) < 0)
    {
      put_string ("ARP failed.\n");
      for (;;);
    }

#if defined(PROXY)
  inet_aton ("10.145.105.245", &server_ip_addr); /* Your proxy server */
#else
  if (dns_resolv (p, &eth_addr, dns_server, server_name, &server_ip_addr) < 0)
    {
      put_string ("DNS failed.\n");
      for (;;);
    }
  put_string ("DNS resolved.\n");
#endif

  if ((netmask & ip_addr) == (netmask & server_ip_addr))
    server = server_ip_addr;
  else
    server = router_ip_addr;
  if (arp_resolv (p, server, &eth_addr) < 1)
    {
      put_string ("ARP failed.\n");
      for (;;);
    }

  put_string ("Address resolution (IP -> Ether MAC) success.\n");

  if (ping (p, &eth_addr, server, 5) == 1)
    put_string ("!\n");

#else /* IPv6 */
  eth_mc_setup_ipv6 ();

  if (net6_address_autoconfig (p) < 0)
    {
      put_string ("*** IPv6 initialization failed.\n");
      for (;;);
    }

  put_string ("My IPv6 address: ");
  put_hex (node_ipv6_addr.bits[0]);
  put_hex (node_ipv6_addr.bits[1]);
  for (i = 1; i < 8; i++)
    {
      put_string (" :");
      put_hex (node_ipv6_addr.bits[i*2]);
      put_hex (node_ipv6_addr.bits[i*2+1]);
    }
  put_string ("\n");

  memcpy (ipv6_note.bits, node_ipv6_addr.bits, 8);

  put_string ("Ping to the notebook.\n");
  if ((dst_eth_addr = ipv6_get_route (p, &ipv6_note)) == 0)
    put_string ("*** No route to the notebook.\n");
  else
    if (ping6 (p, dst_eth_addr, &ipv6_note, 5) == 1)
      put_string ("Got reply!\n");

  put_string ("Ping to the router.\n");
  if ((dst_eth_addr = ipv6_get_route (p, &ipv6_router)) == 0)
    put_string ("*** No route to the router.\n");
  else
    if (ping6 (p, dst_eth_addr, &ipv6_router, 5) == 1)
      put_string ("Got reply!\n");

  put_string ("Ping to the server.\n");
  if ((dst_eth_addr = ipv6_get_route (p, &ipv6_server)) == 0)
    put_string ("*** No route to server.\n");
  else
    if (ping6 (p, dst_eth_addr, &ipv6_server, 5) == 1)
      put_string ("Got reply!\n");

  if (dhcp6 (p, &dns_server) < 0)
    put_string ("*** No DNS server, using hard-coded one.\n");
  else
    put_string ("Got info for DNS by DHCPv6.\n");

  put_string ("Query to DNS server.\n");
  if ((dst_eth_addr = ipv6_get_route (p, &dns_server)) == 0)
    put_string ("*** No route to DNS server, using hard-coded IPv6 address for server...\n");
  else
    if (dns_resolv6 (p, dst_eth_addr, &dns_server, server_name, &ipv6_server) < 0)
      put_string ("DNS failed, using hard-coded IPv6 address for server...\n");
  put_string ("Server IPv6 address: ");
  put_hex (ipv6_server.bits[0]);
  put_hex (ipv6_server.bits[1]);
  for (i = 1; i < 8; i++)
    {
      put_string (" :");
      put_hex (ipv6_server.bits[i*2]);
      put_hex (ipv6_server.bits[i*2+1]);
    }
  put_string ("\n");
#endif

  s = http_request;
  memcpy (s, "GET ", 4);
  s += 4;
#if defined(PROXY)
  memcpy (s, "http://", 7);
  s += 7;
  memcpy (s, server_name, strlen (server_name));
  s += strlen (server_name);
  memcpy (s, "/", 1);
  s += 1;
#endif
  memcpy (s, path, strlen (path));
  s += strlen (path);
#if defined(PROXY)
  memcpy (s, " HTTP/1.0\r\n\r\n", 14); /* Including NUL */
#else
  memcpy (s, " HTTP/1.0\r\nHost: ", 17);
  s += 17;
  memcpy (s, server_name, strlen (server_name));
  s += strlen (server_name);
  memcpy (s, "\r\n\r\n", 5); /* Including NUL */
#endif

#if defined(PROXY)
  server_port = 9515;		/* Your proxy's port */
#else
  server_port = 80;
#endif

  put_string ("Loading kernel from network: ");
#ifdef NET_BOOT_IPv4
  r = tcp_maton (p, &eth_addr, server_ip_addr, server_port, 16786,
		 http_request, strlen (http_request),
		 recv_buf, 0x1000000 /* 4Mbyte */);
#else /* IPv6 */
  if ((dst_eth_addr = ipv6_get_route (p, &ipv6_server)) == 0)
    {
      put_string ("*** No route.\n");
      for (;;);
    }

  r = tcp_maton6 (p, dst_eth_addr, &ipv6_server,
		  server_port, 16793,
		  http_request, strlen (http_request),
		  recv_buf, 0x1000000 /* 4Mbyte */);
#endif

  if (r > 0)			/* Successfully received */
    {
      int i;

      /* XXX: Check status (HTTP/1.? 200 OK) */

      for (i=0; i<r; i++)
	if (recv_buf[i] == '\n' && recv_buf[i+1] == '\r' && recv_buf[i+2] == '\n')
	  {
	    i = i+3;
	    break;
	  }
	else if (recv_buf[i] == '\n' && recv_buf[i+1] == '\n')
	  {
	    i = i+2;
	    break;
	  }

      if (i>=r)
	{
	  put_string ("Bad file loaded.\n");
	  for (;;);
	}

      if ((unsigned int)(recv_buf + i) & 0x3)
	{		/* Assumes that header is larger than 3-byte long. */
	  buf = (unsigned char *)((unsigned int)(recv_buf + i) & ~3);
	  memcpy (buf, recv_buf+i, r-i);
	}
      else			/* We're lucky! */
	buf = (recv_buf + i);

      put_string ("\n");
      flush_cache (0);
      led_out (data);

      goto_kernel ((unsigned long)bios_call_vector,
		   (unsigned char *)MEMORY_START, MEMORY_SIZE, 0, 0, buf);
    }
  else
    {
      put_string ("Failed.\n");
      for (;;);
    }
#endif

#ifdef ATA_BOOT
  if (ata_init () < 0)
    {
      put_string ("*** ATA initialization failed.\n");
      for (;;);
    }

  ata_read_sectors (0, 0, buf, 1);
#endif

#if 0
  put_string ("----------------------------\n");
  {
    int i, j;

    for (i = 0; i < 32; i++)
      {
	put_hex (i*16);
	put_char (':');
	for (j = 0; j < 16; j++)
	  put_hex (buf[i*16+j]);
	put_string ("\n");
      }
  }
  put_string ("----------------------------\n");
#endif

  flush_cache (0);
  led_out (data);
  mdelay (500);
#if 0
  put_string ("Launch to MBR!\n");
#endif

  g00ff ();

  /* Never reached */
  for (;;);
}
