/* 
 * Copyright (C) 2004  Free Software Initiative of Japan
 *
 * This file is subject to the terms and conditions of the GNU Lesser
 * General Public License version 2.1 (or later).
 *
 */

#include "io.h"
#include "delay.h"
#include "conf-serial.h"

/*
#define SIO_STS_RXSC	0x0200
#define SIO_STS_TXSC	0x0100
#define SIO_STS_RXCP	0x0004
#define SIO_STS_TXCP	0x0002
 */

#define SIO_STS_TEMP	0x0001

void
serial_init (void)
{
  outw (0x0300, SIO_CR);	/* Reset TX and RX, Disable TX and RX */
#if 1	/* Dip switch setting would be good */
  outw (0x0080, SIO_MOD0);	/* ignore CTS.  Boot it up anyway */
#else
  outw (0x00C0, SIO_MOD0);	/* CTS/RTS flow control enabled */
#endif
  outw (MOD1_INIT, SIO_MOD1);	/* 8-bit, LSB first */
  outw (0x0000, SIO_INTCR);	/* Disable RX/TX interrupts */
  outw (BAUR_INIT, SIO_BAUR);	/* Set baudrate */
#ifdef RBAUR_INIT
  outw (RBAUR_INIT, SIO_RBAUR);
#endif

  outw (0x0003, SIO_CR);	/* Enable TX and RX */
}

int
put_char (unsigned int c)
{
  unsigned short status;

  do				/* Wait for TX ready */
    {
      status = inw (SIO_STS);
      tiny_delay ();		/* NEVER remove this, it's needed! */
    }
  while ((status & SIO_STS_TEMP) == 0);

  outw ((c & 0xff), SIO_TXB);
  return 0;
}

int
put_string (unsigned char *s)
{
  unsigned char c;
  int r;

  while ((c = *s++))
    {
      if (c == '\n')		/* CR LF */
	if ((r = put_char ('\r')) < 0)
	  return r;

      if ((r = put_char (c)) < 0)
	return r;
    }

  return 0;
}

static const unsigned char hex_string[] = "0123456789ABCDEF";

int
put_hex (unsigned char w)
{
  unsigned char c;

  put_char (' ');
  c = hex_string [(w >> 4) & 0x0f];
  put_char (c);
  c = hex_string [w & 0x0f];
  put_char (c);

  return 0;
}
